#! /usr/bin/python3
# -*- coding: UTF-8 -*-
# 15.09.2014

import os, sys, time, threading
from io import StringIO

class FileVisitor(threading.Thread):
    """
    Durchsucht ab dem startDir [default = '.'] alle Verzeichnisse und Dateien
    und filtert Pfade und Dateien nach unterschiedlichen Kriterien:
    - Substring im Pfad
    - Dateierweiterung in Auswahlliste
    - Substring im Dateinamen
    - Größe der Datei
    - letztes Änderungsdatum der Datei

    Dateien, die gefiltert wurden, werden an eine bearbeitende Funktion
    weitergeleitet.

    Die Filter sind Listen (oder Tuple), in denen Strings oder Substrings
    eingetragen werden können.
    Die Substrings müssen die im Pfad, im Dateinamen oder der Dateiextension
    enthalten sein
    - oder eben nicht (negiert = True invertiert die Logik).

    Wird eine Liste als None definiert, dann wird dieser Filter nicht geprüft.

    Zusätzlich kann die Dateigröße und das letzte Änderungsdatum als Filter-
    kriterium dienen.

    Es werden die Anzahl der angetroffenen Directorys und der Files ermittelt.
    """
    def __init__(self,
                path_negiert = 0,
                name_negiert = 0,
                ext_negiert  = 0,
                path_include = None,
                name_include = None,
                ext_include  = None,
                file_ext = None,
                size_min = None,
                size_max = None,
                time_min = None,
                time_max = None,
                callback = None,
                startDir = None,
                f        = None,
                reset    = True,
                queue    = None,
                visitfiles = None,):

        threading.Thread.__init__(self)

        self.path_negiert = path_negiert
        self.name_negiert = name_negiert
        self.ext_negiert  = ext_negiert
        self.name_include = name_include
        self.path_include = path_include
        self.ext_include  = ext_include
        self.f = f

        self.size_min = size_min
        self.size_max = size_max
        self.time_min = time_min
        self.time_max = time_max
        self.startDir = startDir
        self.lenstartDir = len(startDir) + 1
        self.visitfiles = visitfiles
        self.queue      = queue

        self.cnt_dirs        = 0
        self.cnt_dirs_match  = 0
        self.cnt_dirs_enpty  = 0
        self.cnt_files       = 0
        self.cnt_files_match = 0

        self.startDir = startDir
        self.reset = reset
        self.ende = False


    def killThread(self):
        """ Beendet die Suche, sobald das nächste Verzeichnis geöffnet wird """
        self.ende = True


    def run(self):
        if self.reset:
            self.cnt_dirs        = 0
            self.cnt_dirs_match  = 0
            self.cnt_dirs_empty  = 0
            self.cnt_files_match = 0
            self.cnt_files       = 0

        # Einmalige Vorbereitung von Flags für den gesamten Programmablauf
        if self.size_min is not None or self.size_max is not None:
            self.size_check = True
        else: self.size_check = False

        if self.time_min is not None or self.time_max is not None:
            self.time_check = True
        else: self.time_check = False

        # Die aktuelle Systemzeit, auf ganze Sekunden reduziert
        self.systime = int(time.time())

        # Hier gehts los
        for(thisDir, dirsHere, filesHere) in os.walk(self.startDir):
            # Beenden des Walkers
            if self.ende:
                break

            # Zuerst die Directorys filtern
            passed = True

            # Wenn das Verzeichnis keine Dateien enthält,
            # dann erübrigt sich weitere Arbeit
            if not dirsHere and not filesHere:
                self.cnt_dirs_empty += 1
                passed = False

            else:
                self.cnt_dirs += 1

                # wenn ein Filter für die Pfade definiert ist, dann abarbeiten
                if self.path_include is not None:
                    passed = self.path_filter(thisDir)
                    if self.path_negiert:
                        passed = not passed
#                print(thisDir, self.cnt_dirs)

            # Wenn passed is True, dann die Files einzeln prüfen
            if passed:
                self.cnt_dirs_match += 1

                for file in filesHere:
                    passed = True
                    self.filepath = os.path.join(thisDir, file)
                    self.cnt_files += 1

                    if self.ext_include is not None:
                        passed = self.ext_filter(thisDir, file)
                        if self.ext_negiert:
                            passed = not passed

#                    print(file, passed)

                    if self.name_include is not None and passed:
                        passed = self.name_filter(thisDir, file)
                        if self.name_negiert:
                            passed = not passed

#                    print(file, passed)

                    if self.size_check  and passed:
                        passed = self.size_filter(self.filepath)

#                    print(file, passed)

                    if self.time_check and passed:
                        passed = self.time_filter(self.filepath)

                    # die Datei hat alle Filter passiert und wird nun bearbeitet
                    if passed:
                        self.cnt_files_match += 1

                        # nicht ausführen, wenn nur gezählt werden soll
                        if self.visitfiles:
                            # eine beliebig formatierte, aussagekräftige Liste
                            self.fileAttributs()


            # Die queue wird zur Fortschritts-/Aktivitätsanzeige genutzt.
            # Immer wenn ein neues Verzeichnis geöffnet wird, wird ein Datensatz
            # in die Queue geschrieben.
            if self.queue is not None:
                self.queue.put((thisDir, self.cnt_dirs, self.cnt_dirs_match, self.cnt_dirs_empty,
                                    self.cnt_files, self.cnt_files_match))

        # das Ende der Suche ist erreicht
        # Falls ein Abbruch erfolgte:
        if self.ende:
            self.write('###\n\n')
            self.f.write("\nDie Suche wurde abgebrochen, das Ergebnis ist unvollständig !\n\n")

        # Das Ende der Queue markieren durch Einfügen von None
        if self.queue is not None:
            self.queue.put(None)

        # Hier noch die Statistik schreiben
        if self.f is not None:
            self.print_result()


    def print_result(self):
        """ Schreibt eine Zusammenstellung mit Filtereinstellung und
            Anzahl der gefundenen Verzeichnisse und Dateien
            Diese Absatz wird mit "###" eingeleitet,
            Hier endet die Auswertung der Filenamen in einem Script
        """

        if self.time_min:
            t = time.localtime(self.time_min)
            self.time_min = "{:02d}.{:02d}.{} - {:02}:{:02}:{:02}".\
                    format(t[2], t[1], t[0], t[3], t[4], t[5])

        if self.time_max:
            t = time.localtime(self.time_max)
            self.time_max = "{:02d}.{:02d}.{} - {:02}:{:02}:{:02}".\
                    format(t[2], t[1], t[0], t[3], t[4], t[5])

        # In der ersten Zeile der Datei steht der Name des Startverzeichnis,
        # er wird durch '+' '+' begrenzt
        # An der '#' wird das Ende der Fileliste erkannt
        # Der Dateiname/Pfad steht jeweils in den Zeilen ab Pos. 50.
        # Bei Änderungen des Formats muss scripts.extract_files() anpasst werden.

        self.f.write("### - DIESE ZEILE NICHT EDITIEREN\n\n" )
        self.f.write("Zusammenfassung :\n\n")
        self.f.write("Suche ab Verzeichnis        : {}\n\n".format(self.startDir))
        if self.path_include:
            self.f.write("SubStrings im Pfad          : {}  --> {}\n".\
                    format(self.write_state(self.path_negiert),self.path_include ))
        if self.ext_include:
            self.f.write("DateiExtensions             : {}  --> {}\n".\
                    format( self.write_state(self.ext_negiert),self.ext_include ))
        if self.name_include:
            self.f.write("SubStrings im Dateinamen    : {}  --> {}\n".\
                    format(self.write_state(self.name_negiert), self.name_include ))
        if self.size_min:
            self.f.write("Include Dateien kleiner als : {:12}\n".format(self.size_min))
        if self.size_max:
            self.f.write("Include Dateien  größer als : {:12}\n".format(self.size_max))
        if self.time_min:
            self.f.write("Include Dateien   älter als : {:12}\n".format(self.time_min))
        if self.time_max:
            self.f.write("Include Dateien  jünger als : {:12}\n".format(self.time_max))

        self.f.write("\nVerzeichnisse mit Files     : {:12,d}\n".format(self.cnt_dirs))
        self.f.write("gefilterte Verzeichnisse    : {:12,d}\n".format(self.cnt_dirs_match))
        self.f.write("leere Verzeichnisse         : {:12,d}\n".format(self.cnt_dirs_empty))
        self.f.write("\nAnzahl Dateien insgesamt    : {:12,d}\n".format(self.cnt_files))
        self.f.write("gefilterte  Dateien         : {:12,d}\n".format(self.cnt_files_match))


    def write_state(self, state):
        """ Eine kleine Hilfe für die Formatierung der Texte """
        if state: return "Exclude"
        else: return "Include"


    def path_filter(self, dir):
        """
        Prüft, ob ein Eintrag in path_include im Pfad enthalten ist,
        Verzeichnisse, in denen der Substring enthalten ist,
        werden nicht durchsucht.
        Rückgabe    False (keine Dateien durchsuchen)
                    True  (Dateien durchsuchen)
        """
        passed = False
        for p in self.path_include:
            if dir.find(p) >= 0:
                passed = True
        return passed


    def ext_filter(self, path, file):
        """
        Prüft, ob die Dateierweiterung in der Liste "file_ext" enthalten ist,
        nur diese Dateien werden weiter untersucht.
        Dateien mit nicht aufgeführten Extensions werden ignoriert
        """
        name, ext = os.path.splitext(file)

        # ext darf nicht leer sein
        # für eine leere Dateierweiterung muss der Eintrag "   " angelegt werden

        passed = False
        if ext:             # die Dateierweiterung ist nicht leer
            if ext in self.ext_include:
                passed = True
        else:               # Um leere Extension zu filtern, muss der Eintrag ""
                            # vorhanden sein (eine willkürliche Vereinbarung).
            if '.' in self.ext_include:
                passed =  True
        return passed


    def name_filter(self, path, file):
        """
        Prüft, ob ein Substring der Liste "file_exclude" im Dateinamen enthalten ist,
        diese Dateien werden ignoriert.
        """
        name, ext = os.path.splitext(file)

        passed = False
        for i in self.name_include:
            if name.find(i) >= 0:
                passed = True
                break
        return passed


    def size_filter(self, filepath):
        """
        Prüft, ob die Dateigröße größer als "size_min" bzw.
        kleiner als "size_max".
        """
        passed = False
        size = os.path.getsize(filepath)
        if self.size_min is not None and self.size_max is None:
            if size <= self.size_min:
                passed = True

        elif self.size_max is not None and self.size_min is None:
            if size >= self.size_max:
                passed = True

        elif self.size_max is not None and self.size_min is not None:
            if size <= self.size_min and size >= self.size_max:
                passed = True
        return passed


    def time_filter(self, filepath):
        """
        Prüft, ob die letzte Änderung (in Tagen)
        alter  als time_min und
        jünger als time_max ist.
        """
        filetime = int(os.path.getmtime(filepath))
        passed = False

        if self.time_min is not None and self.time_max is None:
            if filetime < self.time_min:
                passed = True

        elif self.time_max is not None and self.time_min is None:
            if filetime > self.time_max:
                passed = True

        elif self.time_max is not None and self.time_max is not None:
            if filetime > self.time_max and filetime < self.time_min:
                passed = True
        return passed


    def fileAttributs(self):
        """
        Stellt eine Zeile mit Daten der gefilterten Datei zusammen.
        Diese Liste dient nicht der Weiterbearbeitung sondern zur Vermittlung
        der Dateiinformationen.
        Diese Datei wird von den Scripten aufgegriffen und der Dateiname
        ab Position 50 isoliert.
        Veränderungen in der Formatierung müssen daher in den Scripten
        berücksichtigt werden.
        """
        size = os.path.getsize(self.filepath)
        t = int(os.path.getmtime(self.filepath))
        t = time.localtime(t)
        date = "{:02d}.{:02d}.{} {:02}:{:02}:{:02}".format(t[2], t[1], t[0], t[3], t[4], t[5])
        self.f.write("{} - Bytes:{:13,d} - File: {}\n".format(date, size, \
                            self.filepath[self.lenstartDir:]))


if __name__ == '__main__':
    print("WALKER.PY: Diese Datei wird nur aus 'walker_gui' aufgerufen !")
